/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.common.binaryfile;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;
import java.util.function.Function;

import org.apache.commons.io.IOUtils;
import org.springframework.core.task.TaskExecutor;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.errorhandling.TechnicalException;
import io.grpc.stub.CallStreamObserver;
import lombok.Builder;
import lombok.extern.log4j.Log4j2;

@Log4j2
public class GrpcBinaryFileServerDownloader<T> {

	static final int CHUNK_SIZE = 255 * 1024;

	static final int END_OF_STREAM = -1;
	static final int NOTHING_READ = 0;

	private final CallStreamObserver<T> callObserver;
	private final Function<ByteString, T> chunkBuilder;
	private final Consumer<OutputStream> downloadConsumer;
	private final TaskExecutor taskExecutor;

	private final byte[] buffer = new byte[CHUNK_SIZE];
	private final AtomicBoolean started = new AtomicBoolean(false);
	private final AtomicBoolean requestFinished = new AtomicBoolean(false);
	private final AtomicReference<TechnicalException> downloadError = new AtomicReference<>();

	private PipedInputStream inputStream;
	private PipedOutputStream outputStream;

	@Builder
	public GrpcBinaryFileServerDownloader(CallStreamObserver<T> callObserver, Function<ByteString, T> chunkBuilder,
			Consumer<OutputStream> downloadConsumer, TaskExecutor taskExecutor) {
		this.callObserver = callObserver;
		this.chunkBuilder = chunkBuilder;
		this.downloadConsumer = downloadConsumer;
		this.taskExecutor = taskExecutor;
	}

	public void start() {
		if (isNotStarted()) {
			doStart();
		}
	}

	boolean isNotStarted() {
		return started.compareAndSet(false, true);
	}

	void doStart() {
		LOG.debug("Starting download.");
		safelySetupStreams();
		taskExecutor.execute(this::startDownload);
		callObserver.setOnReadyHandler(this::sendChunks);
	}

	void safelySetupStreams() {
		try {
			setupStreams();
		} catch (Exception e) {
			closeOutputStream();
			closeInputStream();
			throw new TechnicalException("Error while setting up streams", e);
		}
	}

	void setupStreams() throws IOException {
		outputStream = new PipedOutputStream();
		inputStream = new PipedInputStream(CHUNK_SIZE);
		outputStream.connect(inputStream);
	}

	void startDownload() {
		try {
			doDownload();
		} catch (Exception e) {
			downloadError.set(new TechnicalException("Error while downloading file contents", e));
		} finally {
			closeOutputStream();
		}
	}

	void doDownload() {
		LOG.debug("Downloading file content...");
		downloadConsumer.accept(outputStream);
		LOG.debug("Download completed.");
	}

	synchronized void sendChunks() {
		try {
			doSendChunks();
		} catch (Exception e) {
			handleError(new TechnicalException("Error while sending chunks", e));
		}
	}

	void doSendChunks() throws IOException {
		while (canSendChunks()) {
			processDataFromInputStream();
		}
	}

	boolean canSendChunks() {
		return !requestFinished.get() && callObserver.isReady();
	}

	void processDataFromInputStream() throws IOException {
		var bytesRead = inputStream.read(buffer);
		switch (bytesRead) {
			case END_OF_STREAM:
				finishProcessing();
				break;
			case NOTHING_READ:
				break;
			default:
				sendBytesToCallObserver(bytesRead);
		}
	}

	void sendBytesToCallObserver(int bytesRead) {
		var bytes = ByteString.copyFrom(buffer, 0, bytesRead);
		var chunk = chunkBuilder.apply(bytes);
		callObserver.onNext(chunk);
		LOG.debug("Sent {} bytes", bytesRead);
	}

	void finishProcessing() {
		if (Objects.nonNull(downloadError.get())) {
			throw downloadError.get();
		} else {
			finishRequest();
			callObserver.onCompleted();
		}
	}

	void handleError(TechnicalException e) {
		LOG.debug("Complete download request with error");
		finishRequest();
		throw e;
	}

	private void finishRequest() {
		requestFinished.set(true);
		closeInputStream();
	}

	void closeOutputStream() {
		IOUtils.closeQuietly(outputStream, e -> LOG.error("OutputStream cannot be closed.", e));
	}

	void closeInputStream() {
		IOUtils.closeQuietly(inputStream, e -> LOG.error("InputStream cannot be closed.", e));
	}
}