/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.common.binaryfile;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.function.BiFunction;
import java.util.function.Function;

import org.apache.commons.lang3.RandomUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.common.binaryfile.BinaryFileTestFactory.TestRequestType;
import de.ozgcloud.common.binaryfile.BinaryFileTestFactory.TestResponseType;
import de.ozgcloud.common.binaryfile.GrpcFileUploadUtils.FileSender;
import de.ozgcloud.common.errorhandling.TechnicalException;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.SneakyThrows;

class GrpcFileUploadUtilsTest {

	@InjectMocks
	private GrpcFileUploadUtils service;

	@Mock
	private BiFunction<byte[], Integer, TestRequestType> chunkBuilder;
	@Mock
	private Function<StreamObserver<TestResponseType>, CallStreamObserver<TestRequestType>> reqObserverBuilder;
	@Mock
	private CallStreamObserver<TestRequestType> requestObserver;
	@Mock
	private InputStream inputStream;

	private FileSender<TestRequestType, TestResponseType> fileSender;

	@Mock
	private TestRequestType metaData;

	@BeforeEach
	void init() {
		when(reqObserverBuilder.apply(any())).thenReturn(requestObserver);
		fileSender = spy(GrpcFileUploadUtils.createSender(chunkBuilder, inputStream, reqObserverBuilder));
	}

	@Nested
	class TestCreateFileSender {

		@Test
		void shouldCreateRequestObserver() {
			GrpcFileUploadUtils.createSender(chunkBuilder, inputStream, reqObserverBuilder).send();

			verify(reqObserverBuilder, atLeastOnce()).apply(notNull());
		}
	}

	@Nested
	class TestSendBinaryFile {

		@Captor
		private ArgumentCaptor<Runnable> runnableCaptor;

		@Test
		void shouldReturnSenderWithFuture() {
			var result = fileSender.send();

			assertThat(result).isNotNull().extracting(FileSender::getResultFuture).isNotNull();
		}
	}

	@Nested
	class TestSendNext {

		@BeforeEach
		void initObserver() {
			fileSender.send();
		}

		@Test
		void shouldCallSendMetaData() {
			fileSender.sendNext();

			verify(fileSender).sendMetaData();
		}

		@Test
		void shouldSendNextChunk() {
			fileSender.sendNext();

			verify(fileSender).sendNextChunk();
		}

	}

	@Nested
	class TestSendNextChunk {

		private final byte[] content = RandomUtils.insecure().randomBytes(GrpcFileUploadUtils.CHUNK_SIZE / 2);
		private ByteArrayInputStream byteArrayInputStream = new ByteArrayInputStream(content);

		@Captor
		private ArgumentCaptor<byte[]> chunkCaptor;

		@Nested
		class TestOnDataAvailable {
			@BeforeEach
			void initObserver() {
				fileSender = spy(GrpcFileUploadUtils.createSender(chunkBuilder, byteArrayInputStream, reqObserverBuilder));
				fileSender.send();
			}

			@Test
			void shouldCallSendChunk() {
				fileSender.sendNextChunk();

				verify(fileSender).sendChunk(chunkCaptor.capture(), eq(content.length));
				assertThat(chunkCaptor.getValue()).contains(content);
			}
		}

		@Nested
		class TestOnNoBytesLeftToRead {

			@Nested
			class TestOnCompleteOnFileSent {
				private static final boolean COMPLETE_ON_FILE_SENT = true;

				@BeforeEach
				void initialize() {
					var buffer = new byte[GrpcFileUploadUtils.CHUNK_SIZE];
					byteArrayInputStream.read(buffer, 0, GrpcFileUploadUtils.CHUNK_SIZE);
					fileSender = spy(GrpcFileUploadUtils.createSender(chunkBuilder, byteArrayInputStream, reqObserverBuilder, COMPLETE_ON_FILE_SENT));
					fileSender.send();
				}

				@Test
				void shouldCallOnCompleted() {
					fileSender.sendNextChunk();

					verify(requestObserver).onCompleted();
				}

				@Test
				void shouldNotCallSendChunk() {
					fileSender.sendNextChunk();

					verify(fileSender, never()).sendChunk(any(), anyInt());
				}

				@Test
				@SneakyThrows
				void shouldCallCloseStreams() {
					fileSender.sendNextChunk();

					verify(fileSender).closeStreams();
				}
			}

			@Nested
			class TestOnNotCompleteOnFileSent {
				private static final boolean COMPLETE_ON_FILE_SENT = false;

				@BeforeEach
				void initialize() {
					var buffer = new byte[GrpcFileUploadUtils.CHUNK_SIZE];
					byteArrayInputStream.read(buffer, 0, GrpcFileUploadUtils.CHUNK_SIZE);
					fileSender = spy(GrpcFileUploadUtils.createSender(chunkBuilder, byteArrayInputStream, reqObserverBuilder, COMPLETE_ON_FILE_SENT));
					fileSender.send();
				}

				@Test
				void shouldNotCallOnCompleted() {
					fileSender.sendNextChunk();

					verify(requestObserver, never()).onCompleted();
				}

				@Test
				void shouldCallSendChunk() {
					fileSender.sendNextChunk();

					verify(fileSender).sendChunk(chunkCaptor.capture(), eq(-1));
					assertThat(chunkCaptor.getValue()).isEmpty();
				}

				@Test
				@SneakyThrows
				void shouldCallCloseStreams() {
					fileSender.sendNextChunk();

					verify(fileSender).closeStreams();
				}
			}
		}

	}

	@Nested
	class TestCloseStreams {

		@Test
		@SneakyThrows
		void shouldCloseInputStream() {
			fileSender.send();

			fileSender.closeStreams();

			verify(inputStream).close();
		}
	}

	@Nested
	class TestSendChunk {

		private static final byte[] CHUNK_PART = "ChunkPartContent".getBytes();

		@BeforeEach
		void initObserver() {
			fileSender.send();
		}

		@Test
		void shouldApplyBuildChunk() throws IOException {
			fileSender.sendChunk(CHUNK_PART, 5);

			verify(chunkBuilder).apply(CHUNK_PART, 5);
		}

		@Test
		void shouldCallOnNext() throws IOException {
			fileSender.sendChunk(CHUNK_PART, 5);

			verify(requestObserver).onNext(any());
		}
	}

	@Nested
	class TestSendMetaData {

		@BeforeEach
		void initObserver() {
			fileSender.send();
		}

		@Test
		void shouldNotSendWithoutMetadata() {
			fileSender.sendMetaData();

			verify(requestObserver, never()).onNext(any());
		}

		@Test
		void shouldSendMetadata() {
			fileSender.withMetaData(metaData).sendMetaData();

			verify(requestObserver).onNext(metaData);
		}

		@Test
		void shouldSendMetadataOnlyOnce() {
			fileSender.withMetaData(metaData).sendMetaData();
			fileSender.sendMetaData();

			verify(requestObserver).onNext(metaData);
		}
	}

	@Disabled("unused")
	@Nested
	class TestReadFromStream {
		@Test
		void shouldThrowException() throws IOException {
			doThrow(IOException.class).when(inputStream).read(any(), anyInt(), anyInt());

			assertThatThrownBy(() -> fileSender.readFromStream()).isInstanceOf(TechnicalException.class);
		}
	}

}